@@ -17,6 +17,15 @@
 #include <libpkgconf/stdinc.h>
 #include <libpkgconf/libpkgconf.h>
 
+#ifndef _WIN32
+#include <fcntl.h>    // open
+#include <libgen.h>   // basename/dirname
+#include <sys/stat.h> // lstat, S_ISLNK
+#include <unistd.h>   // close, readlinkat
+
+#include <string.h>
+#endif
+
 /*
  * !doc
  *
@@ -63,6 +72,65 @@ pkg_get_parent_dir(pkgconf_pkg_t *pkg)
 	char buf[PKGCONF_ITEM_SIZE], *pathbuf;
 
 	pkgconf_strlcpy(buf, pkg->filename, sizeof buf);
+#ifndef _WIN32
+	/*
+	 * We want to resolve symlinks, since ${pcfiledir} should point to the
+	 * parent of the file symlinked to.
+	 */
+	struct stat path_stat;
+	while (!lstat(buf, &path_stat) && S_ISLNK(path_stat.st_mode))
+	{
+		/*
+		 * Have to split the path into the dir + file components,
+		 * in order to extract the directory file descriptor.
+		 *
+		 * The nomenclature here uses the
+		 *
+		 *   ln <source> <target>
+		 *
+		 * model.
+		 */
+		char basenamebuf[PKGCONF_ITEM_SIZE];
+		pkgconf_strlcpy(basenamebuf, buf, sizeof(basenamebuf));
+		const char* targetfilename = basename(basenamebuf);
+
+		char dirnamebuf[PKGCONF_ITEM_SIZE];
+		pkgconf_strlcpy(dirnamebuf, buf, sizeof(dirnamebuf));
+		const char* targetdir = dirname(dirnamebuf);
+
+		const int dirfd = open(targetdir, O_DIRECTORY);
+		if (dirfd == -1)
+			break;
+
+		char sourcebuf[PKGCONF_ITEM_SIZE];
+		ssize_t len = readlinkat(dirfd, targetfilename, sourcebuf, sizeof(sourcebuf) - 1);
+		close(dirfd);
+
+		if (len == -1)
+			break;
+		sourcebuf[len] = '\0';
+
+		memset(buf, '\0', sizeof buf);
+		/*
+		 * The logic here can be a bit tricky, so here's a table:
+		 *
+		 *        <source>      |        <target>        |         result
+		 * -----------------------------------------------------------------------
+		 *     /bar (absolute)  |   foo/link (relative)  |         /bar (absolute)
+		 *   ../bar (relative)  |   foo/link (relative)  |   foo/../bar (relative)
+		 *     /bar (absolute)  |  /foo/link (absolute)  |         /bar (absolute)
+		 *   ../bar (relative)  |  /foo/link (absolute)  |  /foo/../bar (relative)
+		 */
+		if ((sourcebuf[0] != '/')        /* absolute path in <source> wins */
+		    && (strcmp(targetdir, "."))) /* do not prepend "." */
+		{
+			pkgconf_strlcat(buf, targetdir, sizeof buf);
+			pkgconf_strlcat(buf, "/", sizeof buf);
+		}
+		pkgconf_strlcat(buf, sourcebuf, sizeof buf);
+	}
+#endif
+
 	pathbuf = strrchr(buf, PKG_DIR_SEP_S);
 	if (pathbuf == NULL)
 		pathbuf = strrchr(buf, '/');
@@ -124,7 +192,15 @@ static void
 pkgconf_pkg_parser_fragment_func(pkgconf_client_t *client, pkgconf_pkg_t *pkg, const char *keyword, const size_t lineno, const ptrdiff_t offset, const char *value)
 {
 	pkgconf_list_t *dest = (pkgconf_list_t *)((char *) pkg + offset);
+
+	/* we patch client-wide sysroot dir and then patch it back when it is overridden */
+	char *sysroot_dir = client->sysroot_dir;
+	char *pkg_sysroot_dir = pkgconf_tuple_find(client, &pkg->vars, "pc_sysrootdir");
+	if (pkg_sysroot_dir != NULL)
+		client->sysroot_dir = pkg_sysroot_dir;
+
 	bool ret = pkgconf_fragment_parse(client, dest, &pkg->vars, value, pkg->flags);
+	client->sysroot_dir = sysroot_dir;
 
 	if (!ret)
 	{
@@ -250,7 +326,7 @@ determine_prefix(const pkgconf_pkg_t *pkg, char *buf, size_t buflen)
 static char *
 convert_path_to_value(const char *path)
 {
-	char *buf = calloc((strlen(path) + 1) * 2, 1);
+	char *buf = calloc(1, (strlen(path) + 1) * 2);
 	char *bptr = buf;
 	const char *i;
 
@@ -419,7 +495,7 @@ pkgconf_pkg_new_from_file(pkgconf_client_t *client, const char *filename, FILE *
 	pkgconf_pkg_t *pkg;
 	char *idptr;
 
-	pkg = calloc(sizeof(pkgconf_pkg_t), 1);
+	pkg = calloc(1, sizeof(pkgconf_pkg_t));
 	pkg->owner = client;
 	pkg->filename = strdup(filename);
 	pkg->pc_filedir = pkg_get_parent_dir(pkg);
@@ -546,6 +622,9 @@ pkgconf_pkg_free(pkgconf_client_t *client, pkgconf_pkg_t *pkg)
 	if (pkg->copyright != NULL)
 		free(pkg->copyright);
 
+	if (pkg->why != NULL)
+		free(pkg->why);
+
 	free(pkg);
 }
 
@@ -772,15 +851,13 @@ pkgconf_pkg_find(pkgconf_client_t *client, const char *name)
 	{
 		if ((f = fopen(name, "r")) != NULL)
 		{
-			pkgconf_pkg_t *pkg;
-
 			PKGCONF_TRACE(client, "%s is a file", name);
 
 			pkg = pkgconf_pkg_new_from_file(client, name, f, 0);
 			if (pkg != NULL)
 			{
 				pkgconf_path_add(pkg->pc_filedir, &client->dir_list, true);
-				return pkg;
+				goto out;
 			}
 		}
 	}
@@ -1398,19 +1475,24 @@ pkgconf_pkg_verify_dependency(pkgconf_client_t *client, pkgconf_dependency_t *pk
 			return NULL;
 		}
 
-		return pkgconf_pkg_scan_providers(client, pkgdep, eflags);
-	}
-
-	if (pkg->id == NULL)
-		pkg->id = strdup(pkgdep->package);
-
-	if (pkgconf_pkg_comparator_impls[pkgdep->compare](pkg->version, pkgdep->version) != true)
-	{
-		if (eflags != NULL)
-			*eflags |= PKGCONF_PKG_ERRF_PACKAGE_VER_MISMATCH;
+		pkg = pkgconf_pkg_scan_providers(client, pkgdep, eflags);
 	}
 	else
-		pkgdep->match = pkgconf_pkg_ref(client, pkg);
+	{
+		if (pkg->id == NULL)
+			pkg->id = strdup(pkgdep->package);
+
+		if (pkgconf_pkg_comparator_impls[pkgdep->compare](pkg->version, pkgdep->version) != true)
+		{
+			if (eflags != NULL)
+				*eflags |= PKGCONF_PKG_ERRF_PACKAGE_VER_MISMATCH;
+		}
+		else
+			pkgdep->match = pkgconf_pkg_ref(client, pkg);
+	}
+
+	if (pkg != NULL && pkg->why == NULL)
+		pkg->why = strdup(pkgdep->package);
 
 	return pkg;
 }
@@ -1448,7 +1530,11 @@ pkgconf_pkg_report_graph_error(pkgconf_client_t *client, pkgconf_pkg_t *parent,
 			client->already_sent_notice = true;
 		}
 
-		pkgconf_error(client, "Package '%s', required by '%s', not found\n", node->package, parent->id);
+		if (parent->flags & PKGCONF_PKG_PROPF_VIRTUAL)
+			pkgconf_error(client, "Package '%s' not found\n", node->package);
+		else
+			pkgconf_error(client, "Package '%s', required by '%s', not found\n", node->package, parent->id);
+
 		pkgconf_audit_log(client, "%s NOT-FOUND\n", node->package);
 	}
 	else if (eflags & PKGCONF_PKG_ERRF_PACKAGE_VER_MISMATCH)
@@ -1479,6 +1565,8 @@ pkgconf_pkg_walk_list(pkgconf_client_t *client,
 	unsigned int eflags = PKGCONF_PKG_ERRF_OK;
 	pkgconf_node_t *node, *next;
 
+	parent->flags |= PKGCONF_PKG_PROPF_ANCESTOR;
+
 	PKGCONF_FOREACH_LIST_ENTRY_SAFE(deplist->head, next, node)
 	{
 		unsigned int eflags_local = PKGCONF_PKG_ERRF_OK;
@@ -1499,10 +1587,8 @@ pkgconf_pkg_walk_list(pkgconf_client_t *client,
 		if (pkgdep == NULL)
 			continue;
 
-		if (pkgdep->serial == client->serial && !(parent->flags & PKGCONF_PKG_PROPF_VIRTUAL))
+		if((pkgdep->flags & PKGCONF_PKG_PROPF_ANCESTOR) != 0)
 		{
-			pkgdep->identifier = ++client->identifier;
-
 			/* In this case we have a circular reference.
 			 * We break that by deleteing the circular node from the
 			 * the list, so that we dont create a situation where
@@ -1530,13 +1616,13 @@ pkgconf_pkg_walk_list(pkgconf_client_t *client,
 
 		pkgconf_audit_log_dependency(client, pkgdep, depnode);
 
-		pkgdep->identifier = ++client->identifier;
-		pkgdep->serial = client->serial;
 		eflags |= pkgconf_pkg_traverse_main(client, pkgdep, func, data, depth - 1, skip_flags);
 next:
 		pkgconf_pkg_unref(client, pkgdep);
 	}
 
+	parent->flags &= ~PKGCONF_PKG_PROPF_ANCESTOR;
+
 	return eflags;
 }
 
@@ -1616,7 +1702,17 @@ pkgconf_pkg_traverse_main(pkgconf_client_t *client,
 	if (maxdepth == 0)
 		return eflags;
 
-	PKGCONF_TRACE(client, "%s: level %d, serial %lu", root->id, maxdepth, client->serial);
+	/* Short-circuit if we have already visited this node.
+	 */
+	if (root->serial == client->serial)
+		return eflags;
+
+	root->serial = client->serial;
+
+	if (root->identifier == 0)
+		root->identifier = ++client->identifier;
+
+	PKGCONF_TRACE(client, "%s: level %d, serial %"PRIu64, root->id, maxdepth, client->serial);
 
 	if ((root->flags & PKGCONF_PKG_PROPF_VIRTUAL) != PKGCONF_PKG_PROPF_VIRTUAL || (client->flags & PKGCONF_PKG_PKGF_SKIP_ROOT_VIRTUAL) != PKGCONF_PKG_PKGF_SKIP_ROOT_VIRTUAL)
 	{
@@ -1624,30 +1720,29 @@ pkgconf_pkg_traverse_main(pkgconf_client_t *client,
 			func(client, root, data);
 	}
 
-	if (!(client->flags & PKGCONF_PKG_PKGF_SKIP_CONFLICTS))
+	if (!(client->flags & PKGCONF_PKG_PKGF_SKIP_CONFLICTS) && root->conflicts.head != NULL)
 	{
+		PKGCONF_TRACE(client, "%s: walking 'Conflicts' list", root->id);
+
 		eflags = pkgconf_pkg_walk_conflicts_list(client, root, &root->conflicts);
 		if (eflags != PKGCONF_PKG_ERRF_OK)
 			return eflags;
 	}
 
-	PKGCONF_TRACE(client, "%s: walking requires list", root->id);
+	PKGCONF_TRACE(client, "%s: walking 'Requires' list", root->id);
 	eflags = pkgconf_pkg_walk_list(client, root, &root->required, func, data, maxdepth, skip_flags);
 	if (eflags != PKGCONF_PKG_ERRF_OK)
 		return eflags;
 
-	if (client->flags & PKGCONF_PKG_PKGF_SEARCH_PRIVATE)
-	{
-		PKGCONF_TRACE(client, "%s: walking requires.private list", root->id);
+	PKGCONF_TRACE(client, "%s: walking 'Requires.private' list", root->id);
 
-		/* XXX: ugly */
-		client->flags |= PKGCONF_PKG_PKGF_ITER_PKG_IS_PRIVATE;
-		eflags = pkgconf_pkg_walk_list(client, root, &root->requires_private, func, data, maxdepth, skip_flags);
-		client->flags &= ~PKGCONF_PKG_PKGF_ITER_PKG_IS_PRIVATE;
+	/* XXX: ugly */
+	client->flags |= PKGCONF_PKG_PKGF_ITER_PKG_IS_PRIVATE;
+	eflags = pkgconf_pkg_walk_list(client, root, &root->requires_private, func, data, maxdepth, skip_flags);
+	client->flags &= ~PKGCONF_PKG_PKGF_ITER_PKG_IS_PRIVATE;
 
-		if (eflags != PKGCONF_PKG_ERRF_OK)
-			return eflags;
-	}
+	if (eflags != PKGCONF_PKG_ERRF_OK)
+		return eflags;
 
 	return eflags;
 }
@@ -1663,6 +1758,9 @@ pkgconf_pkg_traverse(pkgconf_client_t *client,
 	if (root->flags & PKGCONF_PKG_PROPF_VIRTUAL)
 		client->serial++;
 
+	if ((client->flags & PKGCONF_PKG_PKGF_SEARCH_PRIVATE) == 0)
+		skip_flags |= PKGCONF_PKG_DEPF_PRIVATE;
+
 	return pkgconf_pkg_traverse_main(client, root, func, data, maxdepth, skip_flags);
 }
 
@@ -1736,6 +1834,9 @@ pkgconf_pkg_libs_collect(pkgconf_client_t *client, pkgconf_pkg_t *pkg, void *dat
 	pkgconf_list_t *list = data;
 	pkgconf_node_t *node;
 
+	if (!(client->flags & PKGCONF_PKG_PKGF_SEARCH_PRIVATE) && pkg->flags & PKGCONF_PKG_PROPF_VISITED_PRIVATE)
+		return;
+
 	PKGCONF_FOREACH_LIST_ENTRY(pkg->libs.head, node)
 	{
 		pkgconf_fragment_t *frag = node->data;
